package dev.fitko.fitconnect.client.sender.strategies;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.AttachmentCreationException;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.SchemaNotFoundException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.sender.model.AttachmentPayload;
import dev.fitko.fitconnect.client.sender.model.SendableSubmission;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.api.config.SchemaConfig.METADATA_V_1_0_0;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.*;

public class SendNewSubmissionStrategy {

    private static final Logger LOGGER = LoggerFactory.getLogger(SendNewSubmissionStrategy.class);

    private final Sender sender;

    public SendNewSubmissionStrategy(final Sender sender) {
        this.sender = sender;
    }

    public SentSubmission send(final SendableSubmission sendableSubmission) {

        final UUID destinationId = sendableSubmission.getDestinationId();

        try {

            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId);
            final Destination destination = sender.getDestination(destinationId);

            final List<AttachmentPayload> encryptedAttachments = encryptAndHashAttachments(encryptionKey, sendableSubmission.getAttachments());
            final CreateSubmission newSubmission = buildCreateSubmission(sendableSubmission, encryptedAttachments);

            final SubmissionForPickup announcedSubmission = sender.createSubmission(newSubmission);
            final UUID announcedSubmissionId = announcedSubmission.getSubmissionId();

            final var startTimeAttachmentUpload = StopWatch.start();
            uploadAttachments(encryptedAttachments, announcedSubmissionId);
            LOGGER.info("Uploading attachments took {}", StopWatch.stopWithFormattedTime(startTimeAttachmentUpload));

            LOGGER.info("Creating metadata");
            final Metadata metadata = buildMetadata(sendableSubmission, destination, encryptedAttachments);

            final ValidationResult validatedMetadata = sender.validateMetadata(metadata);
            if (validatedMetadata.hasError()) {
                LOGGER.error("Metadata does not match schema", validatedMetadata.getError());
                return null;
            }

            final String encryptedData = sender.encryptBytes(encryptionKey, sendableSubmission.getData().getBytes(StandardCharsets.UTF_8));
            final String encryptedMetadata = sender.encryptObject(encryptionKey, metadata);

            final var startTimeSubmissionUpload = StopWatch.start();
            final Submission submission = sender.sendSubmission(buildSubmitSubmission(announcedSubmissionId, encryptedData, encryptedMetadata));
            LOGGER.info("Uploading submission took {}", StopWatch.stopWithFormattedTime(startTimeSubmissionUpload));

            LOGGER.info("SUCCESSFULLY HANDED IN SUBMISSION !");
            final AuthenticationTags authenticationTags = buildAuthenticationTags(encryptedAttachments, encryptedData, encryptedMetadata);
            return buildSentSubmission(authenticationTags, submission);

        } catch (final EncryptionException e) {
            LOGGER.error("Encrypting submission failed", e);
        } catch (final RestApiException e) {
            LOGGER.error("Sending submission failed", e);
        } catch (final SchemaNotFoundException e) {
            LOGGER.error("Required schema to send valid submission not found", e);
        } catch (final SubmissionNotCreatedException e) {
            LOGGER.error("Failed to announce new submission", e);
        } catch (final KeyNotRetrievedException e) {
            LOGGER.error("Getting encryption key for destination {} failed", destinationId, e);
        } catch (final InvalidKeyException e) {
            LOGGER.error("Key for destination {} is invalid", destinationId, e);
        } catch (final AttachmentCreationException e) {
            LOGGER.error("Reading file failed. Attachment will not be created.", e);
        }

        return null;
    }

    private Metadata buildMetadata(final SendableSubmission sendableSubmission, final Destination destination, final List<AttachmentPayload> encryptedAttachments) {
        final String hashedData = sender.createHash(sendableSubmission.getData().getBytes(StandardCharsets.UTF_8));

        final String dataMimeType = sendableSubmission.getDataMimeType();
        final URI schemaUri = getSubmissionSchemaFromDestination(destination, dataMimeType);
        final Data data = createData(dataMimeType, hashedData, sendableSubmission.getSchemaUri());
        final List<ApiAttachment> attachmentMetadata = encryptedAttachments.stream().map(this::toHashedAttachment).collect(Collectors.toList());

        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(attachmentMetadata);
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(METADATA_V_1_0_0.toString());
        metadata.setContentStructure(contentStructure);

        setOptionalProperties(sendableSubmission, metadata);

        return metadata;
    }

    private void setOptionalProperties(final SendableSubmission sendableSubmission, final Metadata metadata) {
        if(sendableSubmission.getAuthenticationInformation() != null){
            metadata.setAuthenticationInformation(sendableSubmission.getAuthenticationInformation());
        }
        if(sendableSubmission.getReplyChannel() != null){
            metadata.setReplyChannel(sendableSubmission.getReplyChannel());
        }
        if(sendableSubmission.getPaymentInformation() != null){
            metadata.setPaymentInformation(sendableSubmission.getPaymentInformation());
        }
    }

    private Data createData(final String mimeType, final String hashedData, final URI schemaUri) {
        final var hash = new Hash();
        hash.setContent(hashedData);
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(MimeType.fromValue(mimeType));
        submissionSchema.setSchemaUri(schemaUri);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);
        return data;
    }

    private ApiAttachment toHashedAttachment(final AttachmentPayload attachmentPayload) {
        final var attachment = new ApiAttachment();
        attachment.setAttachmentId(attachmentPayload.getAttachmentId());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentPayload.getFileName());
        attachment.setMimeType(attachmentPayload.getMimeType());
        attachment.setDescription(attachmentPayload.getDescription());

        final var hash = new Hash();
        hash.setContent(attachmentPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(hash);
        return attachment;
    }

    private URI getSubmissionSchemaFromDestination(final Destination destination, final String mimeType) {
        return destination.getServices().stream()
                .map(DestinationService::getSubmissionSchemas)
                .flatMap(Collection::stream)
                .filter(schema -> schema.getMimeType().value().equals(mimeType))
                .map(SubmissionSchema::getSchemaUri)
                .findFirst()
                .orElseThrow(() -> new SchemaNotFoundException("Destination does not support data with mime-type " + mimeType));
    }

    private void uploadAttachments(final List<AttachmentPayload> attachmentPayloads, final UUID submissionId) {
        if (attachmentPayloads.isEmpty()) {
            LOGGER.info("No attachments to upload");
        } else {
            LOGGER.info("Uploading {} attachment(s)", attachmentPayloads.size());
            attachmentPayloads.forEach(a -> sender.uploadAttachment(submissionId, a.getAttachmentId(), a.getEncryptedData()));
        }
    }

    private List<AttachmentPayload> encryptAndHashAttachments(final RSAKey encryptionKey, final List<Attachment> attachments) {
        return attachments.stream()
                .filter(Objects::nonNull)
                .map(attachment -> encryptAndHashAttachment(encryptionKey, attachment))
                .collect(Collectors.toList());
    }

    private AttachmentPayload encryptAndHashAttachment(final RSAKey encryptionKey, final Attachment attachment) {
        try {
            final String encryptedAttachment = sender.encryptBytes(encryptionKey, attachment.getDataAsBytes());
            final String hashedBytes = sender.createHash(attachment.getDataAsBytes());
            return AttachmentPayload.builder()
                    .attachmentId(UUID.randomUUID())
                    .hashedData(hashedBytes)
                    .encryptedData(encryptedAttachment)
                    .mimeType(attachment.getMimeType())
                    .fileName(attachment.getFileName())
                    .description(attachment.getDescription())
                    .build();
        } catch (final EncryptionException e) {
            throw new AttachmentCreationException("Attachment '" + attachment.getFileName() + "' could not be encrypted ", e);
        }
    }
}
